# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import click


@click.command()
@click.argument("message", default="Hello Reprostim World!", required=False)
def echo(message):
    """Echo the provided message or the default value."""
    click.echo(message)



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import logging

import click
from click_didyoumean import DYMGroup

from .. import _init_logger
from ..__about__ import __reprostim_name__, __version__

# setup logging
logger = logging.getLogger(__name__)


def print_version(ctx, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(__version__)
    ctx.exit()


# group to provide commands
@click.group(cls=DYMGroup)
@click.version_option(version=__version__, prog_name=__reprostim_name__)
@click.option(
    "-l",
    "--log-level",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    help="Set the logging level, case-insensitive. Default is INFO.",
)
@click.option(
    "-f",
    "--log-format",
    default="%(asctime)s [%(levelname)s] %(message)s",
    help="Set the logging format string. For the pattern details see standard "
    "Python 'logging.Formatter' documentation.",
)
@click.pass_context
def main(ctx, log_level: str, log_format):
    """Command-line interface to run ReproStim tools and services.
    To see help for the specific command, run:

         reprostim COMMAND --help

    e.g. reprostim timesync-stimuli --help
    """
    # some commands require logging to stderr
    log_to_stderr: bool = ctx.invoked_subcommand in ("qr-parse",)
    _init_logger(log_level.upper(), log_format, log_to_stderr)
    logger.debug(f"{__reprostim_name__} v{__version__}")
    logger.debug(f"main(...), command={ctx.invoked_subcommand}")


# Import all CLI commands
from .cmd_detect_noscreen import detect_noscreen  # noqa: E402
from .cmd_echo import echo  # noqa: E402
from .cmd_list_displays import list_displays  # noqa: E402
from .cmd_monitor_displays import monitor_displays  # noqa: E402
from .cmd_qr_parse import qr_parse  # noqa: E402
from .cmd_timesync_stimuli import timesync_stimuli  # noqa: E402

# List all CLI commands to be included in the main group
__all_commands__ = (
    detect_noscreen,
    echo,
    list_displays,
    monitor_displays,
    qr_parse,
    timesync_stimuli,
)

# Register all CLI commands
for cmd in __all_commands__:
    main.add_command(cmd)



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from time import sleep, time

t0 = time()

import importlib.util
import json
import logging
import os
import shutil
import signal
from datetime import datetime
from enum import Enum

import qrcode

from ..__about__ import __version__

# setup logging
logger = logging.getLogger(__name__)
logger.info("reprostim timesync-stimuli script started")


#######################################################
# Constants

MAX_TR_TIMEOUT: float = 4.0


# Enum for the log event names
class EventName(str, Enum):
    STARTED = "started"
    SERIES_BEGIN = "series_begin"
    SERIES_END = "series_end"
    TRIGGER = "trigger"


# Enum for the mode of the script operation
class Mode(str, Enum):
    # Listen for keyboard events to show codes
    EVENT = "event"
    # Show codes at regular intervals
    INTERVAL = "interval"
    # Just play a beep
    BEEP = "beep"
    # List audio/video devices
    DEVICES = "devices"


#######################################################
# Classes


@dataclass
class SeriesData:
    num: int  # series number
    tr_count: int = 0  # trigger events count in the series
    tr_last_time: float = None  # last trigger event time()
    tr_timeout: float = MAX_TR_TIMEOUT  # trigger event max timeout


#######################################################
# Functions


def get_iso_time(t):
    return datetime.fromtimestamp(t).astimezone().isoformat()


def get_output_file_name(
    prefix: str, start_ts: datetime, end_ts: datetime = None
) -> str:
    start_str: str = get_ts_str(start_ts)
    end_str: str = get_ts_str(end_ts) if end_ts else ""
    return f"{prefix}{start_str}--{end_str}.log"


def get_times():
    t = time()
    return t, get_iso_time(t)


def get_ts_str(ts: datetime) -> str:
    ts_format = "%Y.%m.%d-%H.%M.%S.%f"
    return f"{ts.strftime(ts_format)[:-3]}"


def log(f, rec):
    s = json.dumps(rec).rstrip()
    f.write(s + os.linesep)
    logger.debug(f"LOG {s}")


def mkrec(**kwargs):
    t, tstr = get_times()
    kwargs.update(
        {
            "time": t,
            "time_formatted": tstr,
        }
    )
    return kwargs


def safe_remove(file_name: str):
    if file_name:
        if os.path.isfile(file_name):  # Check if it's a file
            try:
                os.remove(file_name)
                logger.debug(f"File {file_name} deleted successfully.")
            except Exception as e:
                logger.error(f"Failed delete file {file_name}: {e}")
        else:
            logger.warning(f"File {file_name} does not exist or is not a valid file.")


def store_audiocode(audio_file: str, audio_data: int, logfn: str):
    # if audio_file exits, copy it to {logfn}audiocode_{audio_data}.wav
    if os.path.isfile(audio_file):
        sfile = f"{os.path.splitext(logfn)[0]}audiocode_{audio_data}.wav"
        shutil.copy(audio_file, sfile)


#######################################################
# Main script code


def do_init(logfn: str) -> bool:
    if os.path.exists(logfn):
        logger.error(f"Log file {logfn} already exists")
        return False
    return True


def do_main(
    mode: Mode,
    logfn: str,
    is_fullscreen: bool,
    win_size: tuple[int, int],
    display: int,
    qr_scale: float,
    audio_codec: str,
    mute: bool,
    ntrials: int,
    duration: float,
    interval: float,
    keep_audiocode: bool,
    out_func=print,
) -> int:
    logger.info("main script started")

    if not importlib.util.find_spec("psychopy"):
        logger.error(
            "Module 'psychopy' not found, re-install environment with [all] extras"
        )
        return -1

    # late psychopy init
    # setup psychopy logs
    from psychopy import logging as pl

    # pl.console.setLevel(pl.NOTSET)
    pl.console.setLevel(pl.DEBUG)

    from psychopy import core, event, visual

    # from psychopy.hardware import keyboard
    # late audio init
    from ..audio.audiocodes import (
        AudioCodec,
        AudioCodeInfo,
        beep,
        list_audio_devices,
        play_audio,
        save_audiocode,
    )

    sys_shutdown: bool = False

    def check_keys(max_wait: float = 0) -> list[str]:
        keys: list[str]
        if max_wait > 0:
            keys = event.waitKeys(maxWait=max_wait)
        else:
            keys = event.getKeys()
        logger.debug(f"keys={keys}")
        return keys if keys else []

    def on_signal(signum, frame):
        logger.info(f"Received system signal: {signum}")
        if signum in [2, 15] :
            nonlocal sys_shutdown
            logger.debug("setting sys_shutdown=True")
            sys_shutdown = True

    def series_begin(series_num: int) -> SeriesData:
        sd = SeriesData(num=series_num, tr_count=0, tr_timeout=MAX_TR_TIMEOUT)
        logger.debug(f"series begin: {sd.num}")
        # log series begin event
        nonlocal f
        log(
            f,
            mkrec(
                event=EventName.SERIES_BEGIN,
                mode=mode,
                logfn=logfn,
                series_num=sd.num,
            ),
        )
        out_func(f"Series [{sd.num}] started")
        return sd

    def series_end(sd: SeriesData) -> SeriesData:
        if sd:
            logger.debug(f"series end: {sd.num}")
            # log series end event
            nonlocal f
            log(
                f,
                mkrec(
                    event=EventName.SERIES_END,
                    mode=mode,
                    logfn=logfn,
                    series_num=sd.num,
                    trigger_count=sd.tr_count,
                ),
            )
            out_func(f"Series [{sd.num}] ended, trigger count: {sd.tr_count}")
        return None

    if mode == Mode.BEEP:
        for _ in range(ntrials):
            beep(interval * 0.5, async_=True)
            sleep(interval)
        return 0

    if mode == Mode.DEVICES:
        list_audio_devices()
        return 0

    # register signal hook
    signal.signal(signal.SIGINT, on_signal)
    signal.signal(signal.SIGTERM, on_signal)


    audio_data: int = 0
    audio_file: str = None
    audio_info: AudioCodeInfo = None
    f = open(logfn, "w")

    win = visual.Window(
        fullscr=is_fullscreen,
        title=f"ReproStim timesync-stimuli v{__version__}",
        name="timesync-stimuli",
        size=win_size,
        screen=display,
    )
    logger.debug(f"win.winHandle class: {type(win.winHandle).__name__}")
    win.winHandle.on_activate = lambda: out_func("Window activated")
    win.winHandle.on_activate()
    win.winHandle.on_deactivate = lambda: out_func("Window deactivated")
    # win.winHandle.set_caption(win.title)
    win.mouseVisible = False  # hides the mouse pointer
    # win.winHandle.activate()

    # log script started event
    log(
        f,
        mkrec(
            event=EventName.STARTED,
            mode=mode,
            logfn=logfn,
            start_time=t0,
            start_time_formatted=get_iso_time(t0),
        ),
    )

    message = visual.TextStim(
        win,
        text="""Waiting for scanner trigger.\nInstructions
            for Participant...""",
    )
    message.draw()

    fixation = visual.TextStim(win, text="+")
    visual.TextStim(win, text="", pos=(0, -0.7), height=0.05)

    win.flip()

    fixation.draw()  # Change properties of existing stim
    win.flip()

    # spd = 0.500  # Stimulus Presentation Duration
    # soa = 6.000  # Stimulus Onset Asynchrony
    # iwt = 5  # Initial Wait Time between scanner trigger and first stimulus

    # stim_images = []
    # stim_names = []
    # keys = []  # None received/expected

    # kb = keyboard.Keyboard()

    t_start = time()
    t_end = t_start + duration if duration > 0 else None

    logger.debug(f"warming time: {(t_start-t0):.6f} sec")
    logger.debug(f"starting loop with {ntrials} trials...")

    terminate: bool = False

    series_num: int = 0
    cur_series: SeriesData = None

    for acqNum in range(ntrials):
        # check the duration time limit
        if t_end and time() > t_end:
            out_func("Time is up!")
            terminate = True

        # exit processing loop if requested
        if terminate:
            break

        logger.debug(f"trial {acqNum}")

        # prepare audio code file if any
        if not mute:
            if keep_audiocode and audio_file:
                store_audiocode(audio_file, audio_data, logfn)
            safe_remove(audio_file)
            audio_data = acqNum
            audio_file, audio_info_ = save_audiocode(
                code_uint16=audio_data, codec=audio_codec
            )
            audio_info = audio_info_
            logger.debug(f"  {audio_info}")

        if mode == Mode.EVENT:
            out_func("Waiting for an event...")
            while True:
                if sys_shutdown:
                    break

                # check the duration time limit
                if t_end and time() > t_end:
                    out_func("Time is up!")
                    terminate = True
                    break
                keys = check_keys(0.2)
                # only break if 5 or exit keys are pressed
                if (
                    keys
                    and len(keys) > 0
                    and (
                        "q" in keys
                        or "escape" in keys
                        or "5" in keys
                        or "num_5" in keys
                    )
                ):
                    break
                # additional check the series if any expired
                # inside the trigger pulse waiting loop
                if (
                    cur_series
                    and cur_series.tr_count > 0
                    and (time() - cur_series.tr_last_time) > cur_series.tr_timeout
                ):
                    logger.debug("series expired")
                    cur_series = series_end(cur_series)

            # break external loop/terminate
            if terminate:
                break

            if sys_shutdown:
                out_func("Shutdown timesync-stimuli...")
                break

            if cur_series and cur_series.tr_count > 0:
                # calculate time delta from last impulse if any
                dt: float = time() - cur_series.tr_last_time

                if dt > cur_series.tr_timeout:
                    logger.debug(f"timed out after {dt} sec, renew series")
                    cur_series = series_end(cur_series)
                else:
                    # after receiving two first consecutive triggers
                    # pulses -- take the temporal distance between
                    # them +50% as the next value of tr_timeout
                    if cur_series.tr_count == 1:
                        cur_series.tr_timeout = dt * 1.5
                        logger.debug(f"update tr_timeout: {cur_series.tr_timeout}")

        elif mode == Mode.INTERVAL:
            # keys = kb.getKeys(waitRelease=False)
            keys = check_keys()

            target_time = t_start + acqNum * interval
            to_wait = target_time - time()
            # sleep some part of it if long enough
            if to_wait >= 0.2:
                sleep(to_wait * 0.7)
            # busy loop without sleep to not miss it
            while time() < target_time:
                sleep(0)  # pass CPU to other threads
                if sys_shutdown:
                    break
            # 2nd break
            if sys_shutdown:
                out_func("Shutdown timesync-stimuli...")
                break
        else:
            raise ValueError(mode)

        # start series if not started
        if not cur_series:
            cur_series = series_begin(series_num)
            series_num += 1

        # trigger record,
        rec = mkrec(event=EventName.TRIGGER, mode=mode, logfn=logfn, acqNum=acqNum)

        if cur_series:
            rec["series_num"] = cur_series.num

        if mode == Mode.INTERVAL:
            rec["interval"] = interval

        if not mute:
            play_audio(audio_file, async_=True)
            audio_time, audio_time_str = get_times()
            rec["a_time"] = audio_time
            rec["a_time_str"] = audio_time_str
            rec["a_data"] = audio_data
            rec["a_codec"] = audio_codec
            rec["a_f0"] = audio_info.f0
            rec["a_f1"] = audio_info.f1
            rec["a_pre_delay"] = audio_info.pre_delay
            rec["a_post_delay"] = audio_info.post_delay
            rec["a_duration"] = audio_info.duration
            if audio_codec == AudioCodec.NFE:
                rec["a_freq"] = audio_info.nfe_freq
                rec["a_df"] = audio_info.nfe_df

            # NOTE: should we add codec info to the log?
            # like f0, f1, sampleRate, bit_duration, duration, etc

        rec["keys"] = keys
        tkeys, tkeys_str = get_times()
        rec["keys_time"] = tkeys
        rec["keys_time_str"] = tkeys_str
        qr = visual.ImageStim(win, qrcode.make(json.dumps(rec)), pos=(0, 0))
        qr.size = qr.size * qr_scale
        qr.draw()
        win.flip()
        tflip, tflip_str = get_times()
        rec["time_flip"] = tflip
        rec["time_flip_formatted"] = tflip_str
        core.wait(0.5)
        fixation.draw()
        win.flip()
        toff, toff_str = get_times()
        rec["prior_time_off"] = toff
        rec["prior_time_off_str"] = toff_str
        log(f, rec)
        out_func(
            f"Trigger pulse: acq={acqNum}, "
            f"series={cur_series.num if cur_series else 'N/A'}, "
            f"keys={keys}"
        )
        # update trigger event series data
        if cur_series:
            cur_series.tr_count = cur_series.tr_count + 1
            cur_series.tr_last_time = tkeys
        if "q" in keys or "escape" in keys:
            break

    cur_series = series_end(cur_series)

    f.close()

    if keep_audiocode and audio_file:
        store_audiocode(audio_file, audio_data, logfn)
    # cleanup temporary audio code file if any
    safe_remove(audio_file)
    logger.info("main script finished")
    return 0



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT
import logging
import os
import sys

# setup logging
root_logger = logging.getLogger(__name__)
root_logger.setLevel(logging.NOTSET)


def _init_logger(
    log_level: str,
    log_format: str = "%(asctime)s [%(levelname)s] %(message)s",
    _to_stderr: bool = False,
):
    """Initialize logger with the specified log level."""

    handler = None
    # optionally send all log messages to stderr
    if _to_stderr:
        handler = logging.StreamHandler(sys.stderr)
    else:
        handler = logging.StreamHandler(sys.stdout)

    if handler:
        if log_format:
            formatter = logging.Formatter(log_format)
            handler.setFormatter(formatter)
        logging.getLogger().addHandler(handler)

    # set the logging level
    if log_level:
        root_logger.setLevel(log_level)
        # setup log level as environment variable
        # used to setup psychopy logs
        os.environ["REPROSTIM_LOG_LEVEL"] = log_level

    root_logger.debug(f"Logging level set to: {log_level}")



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import logging
from datetime import datetime

import click

# setup logging
logger = logging.getLogger(__name__)


@click.command(help="List information about available GUI/displays.")
@click.option(
    "-p",
    "--provider",
    default="platform",
    type=click.Choice(["platform", "pygame", "pyglet", "pyudev", "quartz", "randr"]),
    help="Set display monitoring providers to be used",
)
@click.option(
    "-f",
    "--format",
    "fmt",
    default="jsonl",
    type=click.Choice(["jsonl", "text"], case_sensitive=False),
    help="Set the output format",
)
@click.pass_context
def list_displays(ctx, provider: str, fmt: str):
    """List information about available GUI/displays."""
    logger.debug("reprostim list-displays")

    from ..qr.disp_mon import DmProvider, do_list_displays

    res: int = 0
    logger.debug("reprostim list-displays script started")

    do_list_displays(DmProvider(provider), fmt, click.echo)

    logger.debug(f"reprostim list-displays script finished: {res}")
    logger.debug(f"Exit on   : {datetime.now()}")
    logger.debug(f"Exit code : {res}")
    return res



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import logging
from datetime import datetime

import click

# setup logging
logger = logging.getLogger(__name__)


@click.command(help="Monitor GUI/displays connection status.")
@click.option(
    "-p",
    "--provider",
    default="platform",
    type=click.Choice(["platform", "pygame", "pyglet", "pyudev", "quartz", "randr"]),
    help="Set display monitoring providers to be used",
)
@click.option(
    "-t",
    "--poll-interval",
    default=60,
    type=int,
    help="Specifies displays check/poll interval in seconds "
    "for providers w/o system callbacks.",
)
@click.option(
    "-w",
    "--max-wait",
    default=3,
    type=int,
    help="Specifies maximum wait interval in seconds, " "or -1 to wait forever.",
)
@click.option(
    "-n",
    "--name",
    default="*",
    type=str,
    help="Specifies display name pattern to be monitored"
    " in Unix shell-style wildcards syntax.",
)
@click.option(
    "-i",
    "--id",
    "d_id",
    default="*",
    type=str,
    help="Specifies display ID pattern to be monitored"
    " in Unix shell-style wildcards syntax.",
)
@click.option(
    "-u",
    "--on-change",
    default=None,
    type=str,
    help="Specifies shell script to be executed when target "
    "display is connected, disconnected or monitor "
    "resolution/mode updated.",
)
@click.option(
    "-e",
    "--ext-proc-command",
    default=None,
    type=str,
    help="Specifies external process or shell script to be "
         "executed when target display is connected. "
         "Note: subprocess is automatically terminated when "
         "display is disconnected.",
)
@click.pass_context
def monitor_displays(
    ctx,
    provider: str,
    poll_interval: int,
    max_wait: int,
    name: str,
    d_id: str,
    on_change: str,
    ext_proc_command: str,
):
    """Monitor GUI/displays connection status."""
    logger.debug("reprostim monitor-displays")

    from ..qr.disp_mon import DmProvider, do_monitor_displays

    res: int = 0
    logger.debug("reprostim monitor-displays script started")

    do_monitor_displays(
        DmProvider(provider),
        poll_interval,
        max_wait,
        name,
        d_id,
        on_change,
        ext_proc_command,
    )

    logger.debug(f"reprostim monitor-displays script finished: {res}")
    logger.debug(f"Exit on   : {datetime.now()}")
    logger.debug(f"Exit code : {res}")
    return res



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import logging
import os
from datetime import datetime

import click

# setup logging
logger = logging.getLogger(__name__)


@click.command(help="PsychoPy reprostim timesync-stimuli script.")
@click.option(
    "-m",
    "--mode",
    type=click.Choice(["event", "interval", "beep", "devices"], case_sensitive=False),
    default="event",
    help="Mode of operation: 'event', 'interval', or 'beep', (default: event).",
)
@click.option(
    "-o",
    "--output-prefix",
    default="output_",
    type=str,
    help="Output log file name prefix.",
)
@click.option(
    "-w",
    "--windowed",
    is_flag=True,
    default=False,
    help="Run script in windowed mode, by default fullscreen mode is used.",
)
@click.option(
    "-z",
    "--size",
    "win_size",
    type=(int, int),
    default=(1920, 1080),
    help="Specify window size as a tuple of integers (default: 1920 1080)."
    "Used only in windowed mode.",
)
@click.option(
    "-d",
    "--display",
    default=1,
    type=int,
    help="Display number as an integer (default: 1).",
)
@click.option(
    "-s",
    "--qr-scale",
    default=0.8,
    type=float,
    help="Specify QR code scale factor in range 0..1. "
    "Use 1.0 to fit full height (default: 0.8).",
)
@click.option(
    "-a",
    "--audio-lib",
    type=click.Choice(
        ["psychopy_sounddevice", "psychopy_ptb", "sounddevice"], case_sensitive=True
    ),
    default="psychopy_sounddevice",
    help="Specify audio library to be used " "(default: psychopy_sounddevice).",
)
@click.option(
    "-c",
    "--audio-codec",
    type=click.Choice(["FSK", "NFE"], case_sensitive=True),
    default="FSK",
    help="Specify audio codec to produce audio code " "(default: FSK).",
)
@click.option(
    "-u",
    "--mute",
    is_flag=True,
    default=False,
    help="Disable audio codes generation (default is False).",
)
@click.option(
    "-t", "--trials", default=300, type=int, help="Specifies number of trials."
)
@click.option(
    "-d",
    "--duration",
    default=-1,
    type=float,
    help="Specifies script duration in seconds. Use negative value "
    "for infinite duration (default: -1).",
)
@click.option(
    "-i",
    "--interval",
    default=2,
    type=float,
    help="Specifies interval value (default: 2.0).",
)
@click.option(
    "-k",
    "--keep-audiocode",
    is_flag=True,
    default=False,
    help="Store audiocode as separate .wav files "
    "for debug purposes (default is False).",
)
@click.pass_context
def timesync_stimuli(
    ctx,
    mode: str,
    output_prefix: str,
    windowed: bool,
    win_size: tuple[int, int],
    display: int,
    qr_scale: float,
    audio_lib: str,
    audio_codec: str,
    mute: bool,
    trials: int,
    duration: float,
    interval: float,
    keep_audiocode: bool,
):
    """Run psychopy script with QR video and audio codes."""
    click.echo("reprostim timesync-stimuli")

    from ..qr.timesync_stimuli import do_init, do_main, get_output_file_name

    is_fullscreen: bool = not windowed

    # psychopy has similar logging levels like
    # default logging module
    # pl.console.setLevel(log_level)
    start_ts: datetime = datetime.now()
    logger.debug("reprostim timesync-stimuli script started")
    logger.debug(f"  Started on   : {start_ts}")
    logger.debug(f"    mode       : {mode}")
    logger.debug(f"    prefix     : {output_prefix}")
    logger.debug(f"    fullscreen : {is_fullscreen}")
    logger.debug(f"    win_size   : {win_size[0]}x{win_size[1]}")
    logger.debug(f"    display    : {display}")
    logger.debug(f"    audio_lib  : {audio_lib}")
    logger.debug(f"    mute       : {mute}")
    logger.debug(f"    duration   : {duration}")
    logger.debug(f"    interval   : {interval}")

    output: str = get_output_file_name(output_prefix, start_ts)
    logger.debug(f"    output     : {output}")

    # setup environment variables
    os.environ["REPROSTIM_AUDIO_LIB"] = audio_lib

    if not do_init(output):
        logger.error()
        return -1

    res = do_main(
        mode,
        output,
        is_fullscreen,
        win_size,
        display,
        qr_scale,
        audio_codec,
        mute,
        trials,
        duration,
        interval,
        keep_audiocode,
        click.echo,
    )

    end_ts: datetime = datetime.now()
    logger.debug(f"  Finished on: {end_ts}")

    # rename log file if any
    output2: str = get_output_file_name(output_prefix, start_ts, end_ts)
    if os.path.exists(output):
        os.rename(output, output2)
        logger.info(f"Output log renamed: {output} -> {output2}")

    logger.info(f"reprostim timesync-stimuli script finished: {res}")
    logger.info(f"Exit on   : {datetime.now()}")
    logger.info(f"Exit code : {res}")
    return res



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT
import fnmatch
import json
import logging
import subprocess
import sys
import time
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from enum import Enum
from itertools import chain
from typing import Generator, Iterable

import psutil

# Display monitoring service: cross-platform API to list
# available GUI displays and monitor status


# initialize the logger
logger = logging.getLogger(__name__)
# logging.getLogger().setLevel(logging.DEBUG)
# logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
logger.debug(f"name={__name__}")


# Platform constants
class DmPlatform(str, Enum):
    LINUX = "linux"
    XOS = "xos"
    # WINDOWS = "win64"


# Service provider constants
class DmProvider(str, Enum):
    """Enum to represent the display monitoring provider/engine
    depending on used OS environment."""

    ALL = "all"  # use all available providers
    PLATFORM = "platform"  # use the best OS specific providers
    PYGAME = "pygame"  # Cross-platform
    PYGLET = "pyglet"  # Cross-platform
    PYUDEV = "pyudev"  # Used in Linux
    QUARTZ = "quartz"  # Used in macOS
    RANDR = "randr"  # Used in Linux


# import necessary provider implementation in runtime based on OS type
if sys.platform.startswith("linux"):
    import pyudev
    from Xlib import display
    from Xlib.ext import randr

    _PLATFORM = DmPlatform.LINUX
elif sys.platform == "darwin":
    import Quartz

    _PLATFORM = DmPlatform.XOS
else:
    logger.error(f"Unsupported OS: {sys.platform}")
    raise NotImplementedError(f"Unsupported OS: {sys.platform}")

logger.debug(f"Use '{_PLATFORM}' platform.")

MAX_DISPLAYS: int = 16


# class representing display info
@dataclass
class DisplayInfo:
    id: str = None
    name: str = None
    width: int = 0
    height: int = 0
    refresh_rate: float = 0.0
    bits_per_pixel: int = 0
    is_connected: bool = False
    is_active: bool = False
    is_main: bool = False
    is_sleeping: bool = False
    provider: DmProvider = None
    ts: str = datetime.now().isoformat()

    # compare display mode
    def eq_mode(self, di) -> bool:
        if di is None:
            return False
        if (
            di.width != self.width
            or di.height != self.height
            or di.refresh_rate != self.refresh_rate
            or di.bits_per_pixel != self.bits_per_pixel
        ):
            return False
        return True

    # Custom equality based on provider+id key
    def eq_key(self, di):
        if isinstance(di, DisplayInfo):
            return self.provider == di.provider and self.id == di.id
        return False


# Display change type
class DisplayChangeType(str, Enum):
    CONNECT = "connect"  # display connected
    DISCONNECT = "disconnect"  # display disconnected
    MODE = "mode"  # mode, resolution or framerate changed


# class representing display change event
@dataclass
class DisplayChangeEvent:
    type: DisplayChangeType = None
    display: DisplayInfo = None
    old_display: DisplayInfo = None
    ts: str = datetime.now().isoformat()


# sample prototype for display change event callback
def display_change_callback(evt: DisplayChangeEvent):
    logger.debug(
        f"display_change_callback: type={evt.type},"
        f" new={evt.display}, old={evt.old_display}"
    )


# pygame implementation
def _enum_displays_pygame() -> Generator[DisplayInfo, None, None]:
    import pygame

    # Initialize Pygame
    pygame.display.init()

    # Get the number of available screens
    disp_count = pygame.display.get_num_displays()
    logger.debug(f"Number of displays connected: {disp_count}")

    # default display info
    def_disp_info = pygame.display.Info()
    def_w = def_disp_info.current_w
    def_h = def_disp_info.current_h
    logger.debug(f"Default display resolution: {def_w}x{def_h}")

    # Loop through each screen and list available modes
    for i in range(disp_count):
        logger.debug(f"Display {i}:")
        di: DisplayInfo = DisplayInfo()
        di.provider = DmProvider.PYGAME
        di.id = i
        di.name = f"display-{i}"

        # logger.debug(f"  Active: {pygame.display.get_active()}")

        disp_w, disp_h = pygame.display.get_desktop_sizes()[i]
        logger.debug(f"  Resolution: {disp_w}x{disp_h}")
        di.is_connected = True
        di.is_active = True
        di.width = disp_w
        di.height = disp_h

        if disp_w == def_w and disp_h == def_h:
            di.is_main = True

        # pygame.display.set_mode((desktop_size[0], desktop_size[1]),
        # pygame.FULLSCREEN, display=i)
        # inf = pygame.display.Info()

        # Get available resolutions for the current display
        # modes = pygame.display.list_modes(display=i)
        # logger.debug(f"Available resolutions on Display {i}:")
        # for mode in modes:
        #    logger.debug(f"{mode[0]}x{mode[1]}")
        yield di

    pygame.display.quit()


# pyglet implementation
def _enum_displays_pyglet() -> Generator[DisplayInfo, None, None]:
    # handle both pyglet versions
    import pyglet

    if hasattr(pyglet, "display"):
        from pyglet.display import get_display
    elif hasattr(pyglet, "canvas"):
        from pyglet.canvas import get_display

    # trace version info
    if hasattr(pyglet, "c"):
        logger.debug(f"pyglet version: {pyglet.c}")

    display = get_display()

    primary_screen = display.get_default_screen()

    # Get all screens connected to the display
    screens = display.get_screens()

    # Enumerate and print screen details
    for i, scr in enumerate(screens):
        # Note: scr has private _cg_display_id property
        # with quartz display ID as well in xos.

        di: DisplayInfo = DisplayInfo()
        di.provider = DmProvider.PYGLET
        di.id = i
        di.name = f"screen-{i}"
        di.is_connected = True
        di.is_active = True

        logger.debug(f"Screen {i}:")
        if hasattr(scr, "get_nsscreen"):
            logger.debug(f"  NSScreen: {scr.get_nsscreen()}")
        if (
            scr.display == primary_screen.display
            and scr.width == primary_screen.width
            and scr.height == primary_screen.height
            and scr.x == primary_screen.x
            and scr.y == primary_screen.y
        ):
            logger.debug("  primary")
            di.is_main = True

        logger.debug(f"  pos  : [{scr.x}, {scr.y}]")
        logger.debug(f"  size : {scr.width}x{scr.height}")
        di.width = scr.width
        di.height = scr.height

        # scr._xinerama = False
        modes = scr.get_modes()
        logger.debug(f"  modes: {modes}")
        # find current mode if any
        for mode in modes:
            if mode.width == scr.width and mode.height == scr.height:
                logger.debug(f"  current mode : {mode}")
                di.refresh_rate = mode.rate
                di.bits_per_pixel = mode.depth

                # fix issue in pyglet/xlib, convert dotclock to refresh rate
                if type(mode).__name__ == "XlibScreenMode":
                    logger.debug("fix rate")
                    if mode.info.htotal != 0 and mode.info.vtotal != 0:
                        di.refresh_rate = (mode.info.dotclock * 1000.0) / (
                            mode.info.htotal * mode.info.vtotal
                        )
                break

        yield di


# pyudev implementation
def _enum_displays_pyudev() -> Generator[DisplayInfo, None, None]:
    ctx = pyudev.Context()

    # Look for devices under the 'drm' subsystem
    for d in ctx.list_devices(subsystem="drm"):
        if "DEVNAME" in d and "ID_PATH" in d:
            logger.debug("Display: ")
            logger.debug(f"  device_node = {d.device_node}")
            logger.debug(f"  ID_PATH     = {d.get('ID_PATH')}")
            for k in d.properties.keys():
                logger.debug(f"  {k} = {d.properties[k]}")
            # logger.debug(f"  props       = {d.properties}")
            di: DisplayInfo = DisplayInfo()
            di.provider = DmProvider.PYUDEV
            # Note: ID_PATH may contain PCI address
            # like pci-0000:01:00.0 or similar,
            # it can be mapped to name by looking at
            #    ls -l /sys/class/drm/
            di.id = d.get("ID_PATH")
            di.name = d.device_node
            # strip prefix to get the name optionally
            if di.name.startswith("/dev/dri/"):
                di.name = di.name[9:]
            di.is_connected = True

            yield di


# randr implementation
def _enum_displays_randr() -> Generator[DisplayInfo, None, None]:
    # Open display connection,
    # TODO: use $DISPLAY data like param e.g. ":0"?
    d = display.Display()
    try:
        root = d.screen().root

        # Get screen resources
        scr_res = randr.get_screen_resources(root)

        # Get the current mode of the primary output
        outputs = scr_res.outputs
        if not outputs:
            logger.debug("No connected outputs found")
            return

        # Iterate over all outputs to find the active one
        for output in outputs:
            di: DisplayInfo = DisplayInfo()
            di.provider = DmProvider.RANDR
            out_info = randr.get_output_info(root, output, scr_res.config_timestamp)
            scr_name = out_info.name  # .decode("utf-8")
            logger.debug(f"Screen name: {scr_name}")
            di.id = scr_name
            di.name = scr_name

            # TODO: check if primary or secondary, e.g vi
            #  xrandr_get_output_primary() or
            # if ??? :
            #     logger.debug("  primary")
            #     di.is_main = True
            # else:
            #     logger.debug("  secondary")
            #     di.is_main = False

            if out_info.connection == 0:
                logger.debug("  connected")
                di.is_connected = True
            else:
                logger.debug("  disconnected")
                di.is_connected = False

            if out_info.crtc == 0:
                logger.debug("  not active")
                di.is_active = False
                yield di
                continue  # Skip outputs that are not active
            else:
                logger.debug("  active")
                di.is_active = True

            crtc_info = randr.get_crtc_info(
                root, out_info.crtc, scr_res.config_timestamp
            )

            # Get max screen width and height
            scr_max_width = crtc_info.width
            scr_max_height = crtc_info.height
            logger.debug(f"  max resolution: {scr_max_width}x{scr_max_height}")

            # Find active mode refresh rate
            for mode in scr_res.modes:
                if mode.id == crtc_info.mode:
                    refresh_rate = round(
                        mode.dot_clock / (mode.h_total * mode.v_total), 2
                    )
                    di.refresh_rate = refresh_rate
                    logger.debug(f"  refresh rate: {refresh_rate} Hz")
                    di.width = mode.width
                    di.height = mode.height
                    logger.debug(f"  resolution: {mode.width}x{mode.height}")
                    break

            yield di
    finally:
        d.close()


# quartz implementation
def _enum_displays_quartz() -> Generator[DisplayInfo, None, None]:
    # logger.debug(dir(Quartz))

    # TODO: find display name,
    #    see also NSScreen.screens @NSScreenNumber to match
    #    with display_id (CGDirectDisplayID)
    #    [https://stackoverflow.com/questions/1236498/how-to-
    #    get-the-display-name-with-the-display-id-in-mac-os-x]
    #    https://github.com/johnjohndoe/DisplayInfo
    #
    # from serial.tools.list_ports_osx import (GetIOServicesByType,
    #                                          IORegistryEntryGetName,
    #                                          get_string_property,
    #                                          get_int_property)
    #
    # services = GetIOServicesByType("IODisplayConnect")
    # logger.debug(f"services={services}")
    # for dev_type in services:
    #    logger.debug(f"dev_type={dev_type}")
    #    den_ename = IORegistryEntryGetName(dev_type)
    #    logger.debug(f"den_ename={den_ename}")
    #    name = get_string_property(dev_type, "AppleDisplay")
    #    logger.debug(f"name={name}")

    main_id = Quartz.CGMainDisplayID()
    logger.debug(f"main_id={main_id}")

    (e, lst_ids, c) = Quartz.CGGetOnlineDisplayList(MAX_DISPLAYS, None, None)
    if e:
        logger.error(f"Quartz.CGGetOnlineDisplayList() failed: {e}")
        return

    logger.debug(f"online displays count : {c}")
    logger.debug(f"online displays: {lst_ids}")

    for d_id in lst_ids:
        logger.debug(f"display id={d_id}")
        di: DisplayInfo = DisplayInfo()
        di.provider = DmProvider.QUARTZ
        di.id = d_id
        vendor_id = Quartz.CGDisplayVendorNumber(d_id)
        model_id = Quartz.CGDisplayModelNumber(d_id)
        serial_id = Quartz.CGDisplaySerialNumber(d_id)
        di.name = f"{vendor_id},{model_id},{serial_id}"  # TODO expand IDs
        di.width = Quartz.CGDisplayPixelsWide(d_id)
        di.height = Quartz.CGDisplayPixelsHigh(d_id)
        di.bits_per_pixel = Quartz.CGDisplayBitsPerPixel(d_id)
        di.is_connected = True
        di.is_active = bool(Quartz.CGDisplayIsActive(d_id))
        di.is_main = d_id == main_id
        di.is_sleeping = bool(Quartz.CGDisplayIsAsleep(d_id))

        # extract additional info, but it should be released
        # dm = c_void_p(Quartz.CGDisplayCopyDisplayMode(d_id))
        dm = Quartz.CGDisplayCopyDisplayMode(d_id)
        if dm:
            try:
                di.refresh_rate = Quartz.CGDisplayModeGetRefreshRate(dm)
                # di.bits_per_pixel = Quartz.CGDisplayModeCopyPixelEncoding(dm)
            finally:
                # TODO: Process finished with exit code 138
                #       (interrupted by signal 10:SIGBUS)
                # Quartz.CGDisplayModeRelease(dm)
                pass

        yield di


def enum_displays(
    provider: DmProvider = DmProvider.PLATFORM,
) -> Generator[DisplayInfo, None, None]:
    a = []
    if provider in (DmProvider.PLATFORM, DmProvider.ALL):
        if _PLATFORM == DmPlatform.LINUX:
            a.append(_enum_displays_pyudev())
            a.append(_enum_displays_randr())
        elif _PLATFORM == DmPlatform.XOS:
            a.append(_enum_displays_quartz())
        else:
            raise NotImplementedError(f"Unsupported platform: {_PLATFORM}")
        if provider == DmProvider.ALL:
            # add all cross-platform providers as well
            a.append(_enum_displays_pygame())
            a.append(_enum_displays_pyglet())
    elif provider == DmProvider.PYUDEV:
        a.append(_enum_displays_pyudev())
    elif provider == DmProvider.QUARTZ:
        a.append(_enum_displays_quartz())
    elif provider == DmProvider.RANDR:
        a.append(_enum_displays_randr())
    elif provider == DmProvider.PYGAME:
        a.append(_enum_displays_pygame())
    elif provider == DmProvider.PYGLET:
        a.append(_enum_displays_pyglet())
    else:
        raise NotImplementedError(f"Unsupported provider: {provider}")
    return chain(*a)


# find matched display info object from provided list
# Note: result can be not the same object as provided
# via di param
def find_di(lst: Iterable[DisplayInfo], di: DisplayInfo) -> DisplayInfo:
    if di is None:
        return None
    for di2 in lst:
        if di.eq_key(di2):
            return di2


def do_list_displays(
    provider: DmProvider = DmProvider.PLATFORM, fmt: str = "jsonl", out_func=print
):
    logger.debug("do_list_displays()")

    last_provider: DmProvider = None
    fmt = fmt.lower()  # make it case-insensitive

    for di in enum_displays(DmProvider(provider)):
        if fmt == "jsonl":
            out_func(json.dumps(asdict(di)))
        elif fmt == "text":
            if last_provider != di.provider:
                last_provider = di.provider
                out_func(f"[{di.provider}]")

            mode: str = (
                f" {di.width}x{di.height}" if di.is_connected and di.width > 0 else ""
            )
            rate: str = (
                f" {round(di.refresh_rate, 2)}Hz"
                if di.is_connected and di.refresh_rate > 0
                else ""
            )
            out_func(
                f"  {di.name} [{di.id}] :{mode}{rate}"
                f"{' connected' if di.is_connected else ' disconnected'}"
                f"{' active' if di.is_active else ''}"
                f"{' primary' if di.is_main else ''}"
            )
        else:
            raise NotImplementedError(f"Unknown format: {format}")


def _kill_process_tree(pid):
    try:
        parent = psutil.Process(pid)
        for child in parent.children(recursive=True):  # Get all child processes
            child.terminate()
        parent.terminate()
    except psutil.NoSuchProcess:
        pass


def _terminate_proc(proc):
    if proc and proc.poll() is None:
        pid = proc.pid
        logger.info(f"terminating external process, pid = {pid}")
        _kill_process_tree(proc.pid)
        return pid
    return None


def do_monitor_displays(
    provider: DmProvider = DmProvider.PLATFORM,
    poll_interval: int = 60,
    max_wait: int = 3,
    name: str = "*",
    d_id: str = "*",
    on_change: str = None,
    ext_proc_command: str = None,
    out_func=print,
):
    logger.debug("do_monitor_displays() enter")

    _eproc = None

    # default callback with event filters and
    # shell commands execution
    def callback_impl(dce: DisplayChangeEvent):
        if name and not fnmatch.fnmatch(str(dce.display.name), name):
            logger.debug(f"event ignored, name mismatch: " f"{name}/{dce.display.name}")
            return

        if d_id and not fnmatch.fnmatch(str(dce.display.id), d_id):
            logger.debug(f"event ignored, id mismatch: " f"{d_id}/{dce.display.id}")
            return

        if on_change and len(on_change) > 0:
            evt_json: str = json.dumps(asdict(dce))
            logger.debug(
                f"execute on_change callback <{on_change}>," f" evt: {evt_json}"
            )
            res = subprocess.run(
                [on_change, evt_json],
                check=False,
                timeout=2 * poll_interval,
                capture_output=True,
                text=True,
            )
            out_func(res.stdout)

        if ext_proc_command and len(ext_proc_command) > 0:
            nonlocal _eproc
            if evt.type == DisplayChangeType.CONNECT:
                if _eproc is not None:
                    out_func(f"Terminating external process, pid={_eproc.pid}")
                    _terminate_proc(_eproc)
                    _eproc = None

                logger.debug("starting external process...")
                try:
                    _eproc = subprocess.Popen(
                        [ext_proc_command], stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    )
                    out_func(
                        f"Started external process, pid={_eproc.pid}, cmd={ext_proc_command}"
                    )
                except Exception as e:
                    logger.error(f"Failed to start external process {ext_proc_command}: {e}")

            if evt.type == DisplayChangeType.DISCONNECT:
                if _eproc is not None:
                    out_func(f"Terminating external process, pid={_eproc.pid}")
                    _terminate_proc(_eproc)
                    _eproc = None

        display_change_callback(dce)

    # callback = display_change_callback
    callback = callback_impl

    max_dt: datetime = (
        datetime.now() + timedelta(seconds=max_wait) if max_wait >= 0 else None
    )

    # build initial display info list
    lst = [di for di in enum_displays(DmProvider(provider))]

    # notify all connected display at startup
    for di in lst:
        if di.is_connected:
            evt: DisplayChangeEvent = DisplayChangeEvent()
            evt.type = DisplayChangeType.CONNECT
            evt.display = di
            callback(evt)

    logger.debug("enter monitor displays cycle")
    seq: int = 1
    try:
        while True:
            logger.debug(f"checking {seq} ...")
            seq += 1
            if max_dt is not None and max_dt < datetime.now():
                logger.debug(f"max_wait({max_wait}) reached")
                break
            time.sleep(poll_interval)
            # get latest display list and compare
            lst2 = [di for di in enum_displays(DmProvider(provider))]

            # find disconnected displays, they exist in
            # old display list, but not exist in new one
            for di in lst:
                di2 = find_di(lst2, di)
                if di2:
                    continue

                # notify only connected displays
                if di.is_connected:
                    evt: DisplayChangeEvent = DisplayChangeEvent()
                    evt.type = DisplayChangeType.DISCONNECT
                    di.is_connected = False
                    evt.display = di
                    callback(evt)

            for di2 in lst2:
                di: DisplayInfo = find_di(lst, di2)

                # check new or just connected displays
                if di is None:
                    # notify only connected displays
                    if di2.is_connected:
                        evt: DisplayChangeEvent = DisplayChangeEvent()
                        evt.type = DisplayChangeType.CONNECT
                        evt.display = di2
                        callback(evt)
                else:
                    # update existing displays
                    evt: DisplayChangeEvent = DisplayChangeEvent()
                    evt.display = di2
                    evt.old_display = di
                    # logger.debug(f"compare di2={di2} and di={di}")
                    if di2.is_connected != di.is_connected:
                        evt.type = (
                            DisplayChangeType.CONNECT
                            if di2.is_connected
                            else DisplayChangeType.DISCONNECT
                        )
                    elif not di2.eq_mode(di):
                        evt.type = DisplayChangeType.MODE

                    if evt.type is not None:
                        callback(evt)

            # store latest display snapshot
            lst = lst2

    finally:
        logger.debug("cleaning up...")
        if _eproc is not None:
            out_func(f"Terminating external process, pid={_eproc.pid}")
            _terminate_proc(_eproc)
            _eproc = None
        logger.debug("cleaned up")

    logger.debug("do_monitor_displays() leave")


if __name__ == "__main__":
    do_list_displays(DmProvider.ALL, "text")
    # do_list_displays(DmProvider.PLATFORM, "text")
    # do_list_displays(DmProvider.PYGAME, "text")
    # do_list_displays(DmProvider.PYGLET, "text")
    do_monitor_displays(
        DmProvider.PLATFORM,
        1,
        60,
        name="*",
        d_id="861767562",
        # d_id="103*",
        on_change="./../../../tools/reprostim-display-on-change",
        ext_proc_command="./../../../tools/reprostim-display-ext-proc-command",
    )



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import logging.config
import subprocess
import time
from datetime import timedelta

import cv2
import numpy as np
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)
logger.debug(f"name={__name__}")


class VideoInfo(BaseModel):
    error: str = Field(None, description="Error message")
    fps: float = Field(..., description="Frames per second (FPS)")
    width: int = Field(..., description="Video frame width")
    height: int = Field(..., description="Video frame height")
    is_invalid_timing: bool = Field(
        False, description="Is video with invalid duration/timing"
    )
    is_truncated: bool = Field(False, description="Is video truncated")
    frames_count: int = Field(..., description="Total number of frames")
    nosignal_count: int = Field(..., description="Total number of nosignal " "frames")
    nosignal_rate: float = Field(
        ..., description="Rate of nosignal frames " "in fraction of total frames"
    )
    scanned_count: int = Field(..., description="Total number of scanned frames")

    def __str__(self):
        return (
            f"VideoInfo({self.width}x{self.height}, fps={self.fps}, "
            f"is_truncated={self.is_truncated}, "
            f"frames_count={self.frames_count}, "
            f"scanned_count={self.scanned_count}, "
            f"nosignal_count={self.nosignal_count}, "
            f"nosignal_rate={self.nosignal_rate})"
        )


# Define the range of colors for the rainbow screen
# This is just an example range, adjust it based on your rainbow screen
lower_rainbow = np.array([0, 50, 50])
upper_rainbow = np.array([30, 255, 255])

# Specify nosignal grid size (8 bands) for custom algorithm
grid_rows: int = 6
grid_cols: int = 8
grid_colors = [[None for _ in range(grid_cols)] for _ in range(grid_rows)]


def auto_fix_video(video_path: str, temp_path: str):
    logger.info(f"Run mediainfo to get video information: mediainfo -i {video_path}")
    res = subprocess.run(
        f"mediainfo -i {video_path}",
        check=True,
        shell=True,
        capture_output=True,
        text=True,
    )
    logger.info(f"[exit code] : {res.returncode}")
    logger.info(f"[stdout]    : {res.stdout}")
    logger.info(f"[stderr]    : {res.stderr}")

    logger.info(f"Run fixup ffmpeg : ffmpeg -i {video_path} -an -c copy {temp_path}")
    res = subprocess.run(
        f"ffmpeg -i {video_path} -c copy {temp_path}",
        check=True,
        shell=True,
        capture_output=True,
        text=True,
    )
    logger.info(f"[exit code] : {res.returncode}")
    logger.info(f"[stdout]    : {res.stdout}")
    logger.info(f"[stderr]    : {res.stderr}")
    logger.info("Video fixup completed.")


# Function to calculate opencv2 color difference
def calc_color_diff(color1, color2):
    b1, g1, r1 = color1.astype(np.int32)
    b2, g2, r2 = color2.astype(np.int32)
    return abs(r1 - r2) + abs(g1 - g2) + abs(b1 - b2)


def has_rainbow(frame):
    # Convert frame to HSV color space
    hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
    # Threshold the frame to extract rainbow colors
    mask = cv2.inRange(hsv, lower_rainbow, upper_rainbow)
    # Count non-zero pixels in the mask
    pixel_count = cv2.countNonZero(mask)
    # logger.debug(f"pixel_count={pixel_count}")
    # If a significant number of rainbow pixels are detected, return True
    return pixel_count > 10000  # Adjust threshold as needed


# match nosignal grid colors using custom algorithm
# based on reference image sample
def has_rainbow2(frame):
    height, width, _ = frame.shape
    cy = height // grid_rows
    cx = width // grid_cols
    n = grid_rows * grid_cols
    diff = 0
    for i in range(grid_rows):
        for j in range(grid_cols):
            x = int(j * cx + cx // 2)
            y = int(i * cy + cy // 2)
            clr1 = grid_colors[i][j]
            clr2 = frame[y, x]
            diff = diff + calc_color_diff(clr1, clr2)

    diff = diff / n
    logger.debug(f"diff={diff}")
    # NOTE: tune the threshold value as needed
    return diff < 35


def init_grid_colors(ref_image_path: str):
    logger.debug(f"ref_image_path={ref_image_path}")

    # Load reference image
    image = cv2.imread(ref_image_path)

    # Get the dimensions of the image
    height, width, _ = image.shape

    # Calculate the height and width of grid cell
    cell_height = height // grid_rows
    cell_width = width // grid_cols

    # Loop over the grid
    for i in range(grid_rows):
        for j in range(grid_cols):
            # Calculate the center of each region
            x = int(j * cell_width + cell_width // 2)
            y = int(i * cell_height + cell_height // 2)

            # Get the pixel color at the center
            # opencv use (row, column) coordinates
            clr = image[y, x]

            # Store the color in the 2D list
            grid_colors[i][j] = clr

    # dump the grid colors
    logger.debug("Nosignal reference grid colors:")
    for i, row in enumerate(grid_colors):
        for j, color in enumerate(row):
            b, g, r = color.astype(np.int32)  # OpenCV stores colors as BGR
            logger.debug(f" grid_colors[{i+1}, {j+1}] : RGB({r}, {g}, {b})")


def find_no_signal(
    video_path: str,
    step: int = 1,
    number_of_checks: int = 0,
    show_progress_sec: float = 0.0,
    check_first_frames: int = 0,
) -> VideoInfo:
    vi: VideoInfo = VideoInfo(
        fps=0,
        width=0,
        height=0,
        is_invalid_timing=False,
        is_truncated=False,
        frames_count=0,
        nosignal_count=0,
        nosignal_rate=0.0,
        scanned_count=0,
    )

    if step < 1:
        vi.error = "Step must be greater than 0"
        return vi

    if check_first_frames > 0 and number_of_checks > 0:
        logger.warning(
            f"check_first_frames value specified, "
            f"ignore number_of_checks({number_of_checks}) value."
        )
        number_of_checks = 0

    if number_of_checks < 0:
        vi.error = "Number of checks must be 0 or greater than 0"
        return vi

    if number_of_checks > 0 and step > 1:
        logger.warning(
            f"Number of checks is set, specified step value({step}) will be ignored."
        )
        step = 1

    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        vi.error = f"Couldn't open the video file: {video_path}"
        return vi

    n1 = cap.get(cv2.CAP_PROP_FRAME_COUNT)
    logger.debug(f"n1={str(n1)}")
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    logger.debug(f"frame_count={str(frame_count)}")
    fps = cap.get(cv2.CAP_PROP_FPS)
    frame_width: int = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height: int = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    duration_sec: float = frame_count / fps if fps > 0 else -1.0
    logger.debug(f"duration_sec={duration_sec}")
    vi.width = frame_width
    vi.height = frame_height
    vi.fps = round(fps, 2)
    logger.info(
        f"Video resolution={frame_width}x{frame_height}, fps={str(fps)}, "
        f"frames count={str(frame_count)}"
    )

    logger.info(
        f"pos ms={cap.get(cv2.CAP_PROP_POS_MSEC)}, "
        f"pos frames={cap.get(cv2.CAP_PROP_POS_FRAMES)}"
    )
    nosignal_frames = []

    #    for i in range(frame_count):
    #        logger.debug(f"i={i}")
    #        ret, frame = cap.read()
    #        if not ret:
    #            logger.error(f"Error reading frame {i}")
    #            break
    #
    #        if not has_rainbow(frame):
    #            no_signal_frames.append(i)

    pos_first_frame: int = 0
    pos_last_frame: int = pos_first_frame + frame_count

    if pos_first_frame > pos_last_frame:
        vi.is_truncated = True
        if check_first_frames == 0:
            vi.error = f"Invalid frame range: {pos_first_frame} - {pos_last_frame}"
            return vi

    if duration_sec < 0 or duration_sec > 2 * 24 * 60 * 60:
        vi.is_invalid_timing = True
        if check_first_frames == 0:
            vi.error = f"Invalid video duration: {duration_sec} seconds"
            return vi

    pos_cur_frame: int = pos_first_frame
    pos_next_frame: int = pos_cur_frame
    nosignal_counter: int = 0
    scan_counter: int = 0
    ts_progress: float = time.time()
    progress_interval: float = show_progress_sec
    while True:
        if show_progress_sec > 0.0 and time.time() > ts_progress:
            dt: str = str(timedelta(milliseconds=int(cap.get(cv2.CAP_PROP_POS_MSEC))))
            logger.info(f"Scanning progress: {pos_cur_frame} / {pos_last_frame}, {dt}")
            ts_progress = time.time() + progress_interval

        logger.debug(
            f"pos_frame={pos_cur_frame}, " f"time={round(pos_cur_frame / fps, 1)}"
        )
        cap.set(cv2.CAP_PROP_POS_FRAMES, pos_cur_frame)
        ret, frame = cap.read()
        if ret is False:
            logger.debug(f"Failed reading frame {pos_cur_frame}")
            break

        # for some rare videos, opencv continues to read frames even
        # after the end of video
        # e.g. Videos/2024/03/2024.03.18.14.39.38.336_2024.03.18.14.44.02.541.mkv
        # to fix this cases, just break the loop
        if pos_cur_frame > pos_last_frame:
            logger.debug(
                f"Failed reading frame sequencer, "
                f"pos_cur_frame={pos_cur_frame} > "
                f"pos_last_frame={pos_last_frame}"
            )
            break

        # also double check number_of_checks for similar cases
        if 0 < number_of_checks <= scan_counter:
            logger.debug(
                f"Failed reading frame, number_of_checks={number_of_checks} "
                f"limit reached"
            )
            break

        scan_counter += 1

        # break if check_first_frames is set and reached
        if check_first_frames > 0 and scan_counter >= check_first_frames:
            logger.debug(f"check_first_frames={check_first_frames} reached")
            break

        # set next frame position
        if number_of_checks > 0:
            pos_next_frame = pos_first_frame + int(
                frame_count * scan_counter / number_of_checks
            )
        else:
            pos_next_frame = pos_cur_frame + step

        if has_rainbow2(frame):
            logger.debug("rainbow-yes")
            nosignal_counter = nosignal_counter + 1
            nosignal_frames.append(pos_cur_frame)
        else:
            logger.debug("rainbow-no")

        pos_cur_frame = pos_next_frame

    logger.debug(f"pos_frame={pos_cur_frame}")
    vi.frames_count = scan_counter if vi.is_truncated else pos_cur_frame
    vi.nosignal_count = nosignal_counter
    vi.scanned_count = scan_counter
    if scan_counter > 0:
        vi.nosignal_rate = round(nosignal_counter / scan_counter, 3)

    cap.release()

    # Calculate time from beginning for each identified frame
    # time_from_beginning = [frame_idx / fps for frame_idx in
    # no_signal_frames]

    return vi



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

try:
    # flake8: noqa: F401
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0"  # Default version

# specify the name of reprostim tool
__reprostim_name__ = "reprostim"



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import logging
import os
import re
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

import click
import cv2
import numpy as np
from pydantic import BaseModel, Field
from pyzbar.pyzbar import ZBarSymbol, decode

# initialize the logger
# Note: all logs out to stderr
logger = logging.getLogger(__name__)
# logging.getLogger().addHandler(logging.StreamHandler(sys.stderr))
logger.debug(f"name={__name__}")


# Define class video info details
class InfoSummary(BaseModel):
    path: Optional[str] = Field(None, description="Video file path")
    rate_mbpm: Optional[float] = Field(
        0.0, description="Video file 'byterate' " "in MB per minute."
    )
    duration_sec: Optional[float] = Field(
        0.0, description="Duration of the video " "in seconds"
    )
    size_mb: Optional[float] = Field(0.0, description="Video file size in MB.")


# Define class for video time info
class VideoTimeInfo(BaseModel):
    success: bool = Field(..., description="Success flag")
    error: Optional[str] = Field(None, description="Error message if any")
    start_time: Optional[datetime] = Field(None, description="Start time of the video")
    end_time: Optional[datetime] = Field(None, description="End time of the video")
    duration_sec: Optional[float] = Field(
        None, description="Duration of the video " "in seconds"
    )


# Define model for parsing summary info
class ParseSummary(BaseModel):
    type: Optional[str] = Field("ParseSummary", description="JSON record type/class")
    qr_count: Optional[int] = Field(0, description="Number of QR codes found")
    parsing_duration: Optional[float] = Field(
        0.0, description="Duration of the " "parsing in seconds"
    )
    # exit code
    exit_code: Optional[int] = Field(-1, description="Number of QR codes found")
    video_full_path: Optional[str] = Field(
        None, description="Full path " "to the video file"
    )
    video_file_name: Optional[str] = Field(
        None, description="Name of the " "video file"
    )
    video_isotime_start: Optional[datetime] = Field(
        None, description="ISO datetime " "video started"
    )
    video_isotime_end: Optional[datetime] = Field(
        None, description="ISO datetime " "video ended"
    )
    video_duration: Optional[float] = Field(
        None, description="Duration of the video " "in seconds"
    )
    video_frame_width: Optional[int] = Field(
        None, description="Width of the " "video frame in px"
    )
    video_frame_height: Optional[int] = Field(
        None, description="Height of the " "video frame in px"
    )
    video_frame_rate: Optional[float] = Field(
        None, description="Frame rate of the " "video in FPS"
    )
    video_frame_count: Optional[int] = Field(
        None, description="Number of frames " "in video file"
    )


# Define the data model for the QR record
class QrRecord(BaseModel):
    type: Optional[str] = Field("QrRecord", description="JSON record type/class")
    index: Optional[int] = Field(
        None, description="Zero-based i    ndex of the QR code"
    )
    frame_start: Optional[int] = Field(
        None, description="Frame number where QR code starts"
    )
    frame_end: Optional[int] = Field(
        None, description="Frame number where QR code ends"
    )
    isotime_start: Optional[datetime] = Field(
        None, description="ISO datetime where QR " "code starts"
    )
    isotime_end: Optional[datetime] = Field(
        None, description="ISO datetime where QR " "code ends"
    )
    time_start: Optional[float] = Field(
        None, description="Position in seconds " "where QR code starts"
    )
    time_end: Optional[float] = Field(
        None, description="Position in seconds " "where QR code ends"
    )
    duration: Optional[float] = Field(
        None, description="Duration of the QR code " "in seconds"
    )
    data: Optional[dict] = Field(None, description="QR code data")

    def __str__(self):
        return (
            f"QrRecord(frames=[{self.frame_start}, {self.frame_end}], "
            f"times=[{self.time_start}, {self.time_end} sec], "
            f"duration={self.duration} sec, "
            f"isotimes=[{self.isotime_start}, {self.isotime_end}], "
            f"data={self.data})"
        )


def calc_time(ts: datetime, pos_sec: float) -> datetime:
    return ts + timedelta(seconds=pos_sec)


def get_iso_time(ts: str) -> datetime:
    dt: datetime = datetime.fromisoformat(ts)
    dt = dt.replace(tzinfo=None)
    return dt


def get_video_time_info(path_video: str) -> VideoTimeInfo:
    res: VideoTimeInfo = VideoTimeInfo(
        success=False, error=None, start_time=None, end_time=None
    )
    # Define the regex pattern for the timestamp and file extension
    # (either .mkv or .mp4)
    pattern1 = (
        r"^(\d{4}\.\d{2}\.\d{2}\.\d{2}\.\d{2}\.\d{2}\.\d{3})"
        r"_(\d{4}\.\d{2}\.\d{2}\.\d{2}\.\d{2}\.\d{2}\.\d{3})\.(mkv|mp4)$"
    )

    # add support for new file format
    pattern2 = (
        r"^(\d{4}\.\d{2}\.\d{2}\-\d{2}\.\d{2}\.\d{2}\.\d{3})"
        r"\-\-(\d{4}\.\d{2}\.\d{2}\-\d{2}\.\d{2}\.\d{2}\.\d{3})\.(mkv|mp4)$"
    )

    file_name: str = os.path.basename(path_video)
    logger.info(f"Video file name  : {file_name}")

    ts_format = ""
    match1: str = re.match(pattern1, file_name)
    if match1:
        # Define the format for datetime parsing
        ts_format = "%Y.%m.%d.%H.%M.%S.%f"
    else:
        match1 = re.match(pattern2, file_name)
        if match1:
            # Define the format for datetime parsing
            ts_format = "%Y.%m.%d-%H.%M.%S.%f"
        else:
            res.error = f"Filename '{path_video}' does not match the required pattern."
            return res

    start_ts, end_ts, extension = match1.groups()

    try:
        # Parse the timestamps
        res.start_time = datetime.strptime(start_ts, ts_format)
        res.end_time = datetime.strptime(end_ts, ts_format)
    except ValueError as e:
        res.error = f"Timestamp parsing error: {e}"
        return res

    # Validate the chronological order
    if res.start_time >= res.end_time:
        res.error = "Start timestamp is not earlier than end timestamp."
        return res

    # calculate the duration in seconds
    dt: float = (res.end_time - res.start_time).total_seconds()
    res.duration_sec = dt

    res.success = True
    return res


def finalize_record(
    ps: ParseSummary, vti: VideoTimeInfo, record: QrRecord, iframe: int, pos_sec: float
) -> QrRecord:
    record.frame_end = iframe
    # Note: unclear should we also use last frame duration or not
    record.isotime_end = calc_time(vti.start_time, pos_sec)
    record.time_end = pos_sec
    record.duration = record.time_end - record.time_start
    record.index = ps.qr_count
    logger.info(f"QR: {str(record)}")
    # dump times
    event_time = get_iso_time(record.data["time_formatted"])
    keys_time = get_iso_time(record.data["keys_time_str"])
    logger.info(f" - QR code isotime : {record.isotime_start}")
    logger.info(
        f" - Event isotime   : "
        f"{event_time} / "
        f"dt={(event_time - record.isotime_start).total_seconds()} sec"
    )
    logger.info(
        f" - Keys isotime    : "
        f"{keys_time} / "
        f"dt={(keys_time - record.isotime_start).total_seconds()} sec"
    )
    # print(record.json())
    ps.qr_count += 1
    return record


def do_info_file(path: str):
    logger.info(f"do_info_file({path})")
    vti: VideoTimeInfo = get_video_time_info(path)
    if not vti.success:
        logger.error(f"Failed parse file name time pattern, error: {vti.error}")
        return
    o: InfoSummary = InfoSummary()
    o.path = path
    o.duration_sec = round(vti.duration_sec, 1)
    size: float = os.path.getsize(path)
    o.size_mb = round(size / (1000 * 1000), 1)
    if o.duration_sec > 0.0001:
        o.rate_mbpm = round(size * 60 / (o.duration_sec * 1000 * 1000), 1)
    return o


def do_info(path: str):
    p = Path(path)
    if p.is_file():
        yield do_info_file(path)
    elif p.is_dir():
        logger.info(f"Processing video directory: {path}")
        for root, _, files in os.walk(path):
            for file in files:
                if file.endswith(".mkv"):
                    yield do_info_file(os.path.join(root, file))
            # Uncomment to visit only top-level dir
            # break
    else:
        logger.error(f"Path not found: {path}")


def do_parse(path_video: str):
    ps: ParseSummary = ParseSummary()

    vti: VideoTimeInfo = get_video_time_info(path_video)
    if not vti.success:
        logger.error(f"Failed parse file name time pattern, error: {vti.error}")
        return

    logger.info(f"Video start time : {vti.start_time}")
    logger.info(f"Video end time   : {vti.end_time}")
    logger.info(f"Video duration   : {vti.duration_sec} sec")

    dt = time.time()
    cap = cv2.VideoCapture(path_video)

    # Check if the video opened successfully
    if not cap.isOpened():
        logger.error("Error: Could not open video.")
        return

    # dump video metadata
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = cap.get(cv2.CAP_PROP_FPS)
    frame_width: int = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height: int = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    duration_sec: float = frame_count / fps if fps > 0 else -1.0
    logger.info("Video media info : ")
    logger.info(f"    - resolution : {frame_width}x{frame_height}")
    logger.info(f"    - frame rate : {str(fps)} FPS")
    logger.info(f"    - duration   : {str(duration_sec)} sec")
    logger.info(f"    - frame count: {str(frame_count)}")
    ps.video_frame_rate = fps
    ps.video_frame_width = frame_width
    ps.video_frame_height = frame_height
    ps.video_frame_count = frame_count
    ps.video_duration = duration_sec
    ps.video_isotime_start = vti.start_time
    ps.video_isotime_end = vti.end_time
    ps.video_full_path = path_video
    ps.video_file_name = os.path.basename(path_video)

    if abs(duration_sec - vti.duration_sec) > 120.0:
        logger.error(
            f"Video duration significant mismatch (real/file name):"
            f" {duration_sec} sec vs {vti.duration_sec} sec"
        )

    # for f in vid.iter_frames(with_times=True):

    # TODO: just use tqdm for progress indication
    iframe: int = 0
    pos_sec: float = 0.0
    record: QrRecord = None

    while True:
        iframe += 1
        # pos time in ms
        pos_sec = round((iframe - 1) / fps, 3)
        ret, frame = cap.read()
        if not ret:
            break

        f = np.mean(frame, axis=2)  # poor man greyscale from RGB

        if np.mod(iframe, 50) == 0:
            logger.debug(f"iframe={iframe} {np.std(f)}")

        #    if np.std(f) > 10:
        #        cv2.imwrite('grayscale_image.png', f)
        #        import pdb; pdb.set_trace()

        cod = decode(f, symbols=[ZBarSymbol.QRCODE])
        if len(cod) > 0:
            logger.debug("Found QR code: " + str(cod))
            assert len(cod) == 1, f"Expecting only one, got {len(cod)}"
            data = eval(eval(str(cod[0].data)).decode("utf-8"))
            if record is not None:
                if data == record.data:
                    # we are still in the same QR code record
                    logger.debug("Same QR code: continue")
                    continue
                # It is a different QR code! we need to finalize current one
                yield finalize_record(ps, vti, record, iframe, pos_sec)
                record = None
            # We just got beginning of the QR code!
            logger.debug("New QR code: " + str(data))
            record = QrRecord()
            record.frame_start = iframe
            record.isotime_start = calc_time(vti.start_time, pos_sec)
            record.isotime_end = None
            record.time_start = pos_sec
            record.data = data
        else:
            if record:
                yield finalize_record(ps, vti, record, iframe, pos_sec)
                record = None

    if record:
        yield finalize_record(ps, vti, record, iframe, pos_sec)
        record = None

    ps.exit_code = 0
    ps.parsing_duration = round(time.time() - dt, 1)
    yield ps
    # print(ps.json())


@click.command(help="Utility to parse video and locate integrated " "QR time codes.")
@click.argument("path", type=click.Path(exists=True))
@click.option(
    "--mode",
    default="PARSE",
    type=click.Choice(["PARSE", "INFO"]),
    help='Specify execution mode. Default is "PARSE", '
    "normal execution. "
    'Use "INFO" to dump video file info like duration, '
    "bitrate, file size etc, (in this case "
    '"path" argument specifies video file or directory '
    "containing video files).",
)
@click.pass_context
def main(ctx, path: str, mode: str):
    logger.debug("qr-parse command")
    logger.debug(f"Working dir      : {os.getcwd()}")
    logger.info(f"Video full path  : {path}")

    if not os.path.exists(path):
        logger.error(f"Path does not exist: {path}")
        return 1

    if mode == "PARSE":
        for item in do_parse(path):
            click.echo(item.model_dump_json())
    elif mode == "INFO":
        for item in do_info(path):
            click.echo(item.model_dump_json())
    else:
        logger.error(f"Unknown mode: {mode}")
        return -1
    return 0



# SPDX-FileCopyrightText: 2020-2025 ReproNim Team <info@repronim.org>
#
# SPDX-License-Identifier: MIT

import importlib
import logging
import os
import tempfile
import time
from datetime import datetime
from enum import Enum

import numpy as np
from reedsolo import RSCodec
from scipy.io import wavfile
from scipy.io.wavfile import read, write

# optionally: import sounddevice as sd
sd = (
    importlib.import_module("sounddevice")
    if importlib.util.find_spec("sounddevice")
    else None
)


# setup logging
logger = logging.getLogger(__name__)
logger.setLevel(os.environ.get("REPROSTIM_LOG_LEVEL", "INFO"))

######################################
# Setup psychopy audio library


# Enum for the audio libs
class AudioLib(str, Enum):
    # PsychoPy SoundDevice audio lib
    PSYCHOPY_SOUNDDEVICE = "psychopy_sounddevice"
    # PsychoPy SoundPTB audio lib
    PSYCHOPY_PTB = "psychopy_ptb"
    # sounddevice audio lib
    # http://python-sounddevice.readthedocs.io/
    SOUNDDEVICE = "sounddevice"


_audio_lib = os.environ.get("REPROSTIM_AUDIO_LIB", AudioLib.PSYCHOPY_SOUNDDEVICE)

# optionally import psychopy
try:
    from psychopy import prefs  # noqa: E402

    prefs.hardware["audioLib"] = ["sounddevice"]
    if _audio_lib == AudioLib.PSYCHOPY_SOUNDDEVICE:
        logger.debug("Set psychopy audio library: sounddevice")
        prefs.hardware["audioLib"] = ["sounddevice"]
    elif _audio_lib == AudioLib.PSYCHOPY_PTB:
        logger.debug("Set psychopy audio library: ptb")
        prefs.hardware["audioLib"] = ["ptb"]

    from psychopy import core, sound  # noqa: E402
    from psychtoolbox import audio  # noqa: E402
except ImportError:
    logger.warn(
        "psychopy module not found, if necessary "
        "install reprostim with [all] extra dependencies"
    )


######################################
# Audio code/qr helper functions


def bit_enumerator(data):
    if isinstance(data, DataMessage):
        data = data.encode()
    if isinstance(data, str):
        # If data is a string, iterate over each character
        for char in data:
            if char not in ("0", "1"):
                raise ValueError("String must only contain '0' and '1'.")
            yield int(char)  # Yield the bit as an integer
    elif isinstance(data, bytes):
        # If data is bytes, iterate over each byte
        for byte in data:
            for i in range(7, -1, -1):  # Iterate from MSB to LSB
                yield (byte >> i) & 1  # Extract and yield the bit
    else:
        raise TypeError(
            "Data must be either a string or bytes. Got: " + str(type(data))
        )


# Convert a list of bits to bytes
def bits_to_bytes(detected_bits):
    # Check if the length of detected_bits is a multiple of 8
    if len(detected_bits) % 8 != 0:
        raise ValueError(
            f"Detected bits array must be aligned to 8 bits. Length={len(detected_bits)}"
        )

    byte_array = bytearray()  # Use bytearray for mutable bytes
    for i in range(0, len(detected_bits), 8):
        # Get the next 8 bits
        byte_bits = detected_bits[i : i + 8]  # noqa: E203

        # Convert the list of bits to a byte (big-endian)
        byte_value = 0
        for bit in byte_bits:
            byte_value = (byte_value << 1) | bit  # Shift left and add the bit

        byte_array.append(byte_value)  # Append the byte to the bytearray

    return bytes(byte_array)  # Convert bytearray to bytes


def crc8(data: bytes, polynomial: int = 0x31, init_value: int = 0x00) -> int:
    crc = init_value
    for byte in data:
        crc ^= byte  # XOR byte into the CRC
        for _ in range(8):  # Process each bit
            if crc & 0x80:  # If the highest bit is set
                crc = (crc << 1) ^ polynomial  # Shift left and XOR with polynomial
            else:
                crc <<= 1  # Just shift left
            crc &= 0xFF  # Keep CRC to 8 bits
    return crc


######################################
# Constants


class AudioCodec(str, Enum):
    # Frequency Shift Keying (FSK) where binary data is
    # encoded as two different frequencies f0 and f1 with
    # a fixed bit duration (baud rate or bit_rate).
    FSK = "FSK"

    # Numerical Frequency Encoding (NFE) numbers are mapped
    # directly to specific frequencies
    # can encode only some numeric hash.
    NFE = "NFE"


######################################
# Classes


# Class representing audio data message in big-endian
# format and encoded with Reed-Solomon error correction
# where the message is structured as follows:
#  - 1st byte is CRC-8 checksum
#  - 2nd byte is the length of the data
#  - 3+ and the rest is the data itself
class DataMessage:
    def __init__(self):
        self.value: bytes = b""
        self.length: int = 0
        self.crc8: int = 0
        self.use_ecc: bool = True
        self.rsc = RSCodec(4)

    def decode(self, data: bytes):
        if self.use_ecc:
            dec, dec_full, errata_pos_all = self.rsc.decode(data)
            data = bytes(dec)

        # logger.debug(f"decoded data  : {data}")
        self.crc8 = data[0]
        self.length = data[1]
        self.value = data[2:]
        n = crc8(self.value)
        if self.crc8 != n:
            raise ValueError(f"CRC-8 checksum mismatch: {self.crc8} <-> {n}")

    def encode(self) -> bytes:
        logger.debug("size info")
        logger.debug(f"  - data      : {len(self.value)} bytes, {self.value}")
        b: bytes = bytes([self.crc8, self.length]) + self.value
        logger.debug(f"  - message   : {len(b)} bytes, {b}")
        if self.use_ecc:
            b = bytes(self.rsc.encode(b))
            logger.debug(f"  - ecc       : {len(b)} bytes, {b}")
        return b

    def get_bytes(self) -> bytes:
        return self.value

    def get_str(self) -> str:
        return self.value.decode("utf-8")

    def get_uint(self) -> int:
        c = len(self.value)
        if c == 2:
            return self.get_uint16()
        elif c == 4:
            return self.get_uint32()
        elif c == 8:
            return self.get_uint64()
        else:
            raise ValueError(f"Data length must be 2, 4, or 8 bytes, " f"but was {c}")

    def get_uint16(self) -> int:
        if len(self.value) != 2:
            raise ValueError(
                f"Data length for uint16 must be 2 bytes, " f"but was {len(self.value)}"
            )
        return int.from_bytes(self.value, "big")

    def get_uint32(self) -> int:
        if len(self.value) != 4:
            raise ValueError(
                f"Data length for uint32 must be 4 bytes, " f"but was {len(self.value)}"
            )
        return int.from_bytes(self.value, "big")

    def get_uint64(self) -> int:
        if len(self.value) != 8:
            raise ValueError(
                f"Data length for uint64 must be 8 bytes, " f"but was {len(self.value)}"
            )
        return int.from_bytes(self.value, "big")

    def set_bytes(self, data: bytes):
        self.value = data
        self.length = len(data)
        self.crc8 = crc8(data)

    def set_str(self, s: str):
        self.set_bytes(s.encode("utf-8"))

    def set_uint16(self, i: int):
        self.set_bytes(i.to_bytes(2, "big"))

    def set_uint32(self, i: int):
        self.set_bytes(i.to_bytes(4, "big"))

    def set_uint64(self, i: int):
        self.set_bytes(i.to_bytes(8, "big"))


# Class to provide general information about audio code
class AudioCodeInfo:
    def __init__(self):
        self.codec = None
        self.f1 = None
        self.f0 = None
        self.nfe_df = None
        self.sample_rate = None
        self.bit_duration = None
        self.nfe_duration = None
        self.nfe_freq = None
        self.bit_count = None
        self.volume = None
        self.duration = None
        self.pre_delay = None
        self.post_delay = None

    # to string
    def __str__(self):
        return (
            f"AudioCodeInfo(codec={self.codec}, "
            f"f1={self.f1}, "
            f"f0={self.f0}, "
            f"nfe_df={self.nfe_df}, "
            f"rate={self.sample_rate}, "
            f"bit_duration={self.bit_duration}, "
            f"bit_count={self.bit_count}, "
            f"nfe_freq={self.nfe_freq}, "
            f"volume={self.volume}, "
            f"duration={self.duration})"
            f"pre_delay={self.pre_delay}, "
            f"post_delay={self.post_delay}"
        )


# Class to generate/parse QR-like audio codes with FSK modulation
class AudioCodeEngine:
    def __init__(
        self,
        codec=AudioCodec.FSK,
        f0=1000,
        f1=5000,
        nfe_df=100,  # used only in NFE
        sample_rate=44100,
        bit_duration=0.0070,  # used only in FSK
        nfe_duration=0.3,  # used only in NFE
        volume=0.80,
        pre_delay=0.1,
        pre_f=0,  # 1780
        post_delay=0.1,
        post_f=0,  # 3571
    ):
        self.codec = codec
        self.f0 = f0
        self.f1 = f1
        self.nfe_df = nfe_df
        self.sample_rate = sample_rate
        self.bit_duration = bit_duration
        self.nfe_duration = nfe_duration
        if volume < 0.0 or volume > 1.0:
            raise ValueError("Volume must be between 0.0 and 1.0.")
        self.volume = volume
        self.pre_delay = pre_delay
        self.pre_f = pre_f
        self.post_delay = post_delay
        self.post_f = post_f

    def generate_sin(self, freq_hz, duration_sec):
        t = np.linspace(
            0, duration_sec, int(self.sample_rate * duration_sec), endpoint=False
        )
        signal = self.volume * np.sin(2 * np.pi * freq_hz * t)
        return signal

    def generate_fsk(self, data) -> (np.array, AudioCodeInfo):
        logger.debug(
            f"audio config  : f1={self.f1} Hz, f0={self.f0} Hz, "
            f"rate={self.sample_rate} Hz, "
            f"bit duration={self.bit_duration} sec, "
            f"volume={self.volume}"
        )
        t = np.linspace(
            0,
            self.bit_duration,
            int(self.sample_rate * self.bit_duration),
            endpoint=False,
        )

        # Create FSK signal
        fsk_signal = np.array([])
        pre_signal = np.array([])
        post_signal = np.array([])

        # generate pre-signal if any
        if self.pre_delay > 0:
            pre_signal = self.generate_sin(self.pre_f, self.pre_delay)

        # generate post-signal if any
        if self.post_delay > 0:
            post_signal = self.generate_sin(self.post_f, self.post_delay)

        # generate data signal properly
        c: int = 0
        sb: str = ""
        for bit in bit_enumerator(data):
            c += 1
            sb += str(bit)
            if bit == 1:
                fsk_signal = np.concatenate(
                    (fsk_signal, self.volume * np.sin(2 * np.pi * self.f1 * t))
                )
            else:
                fsk_signal = np.concatenate(
                    (fsk_signal, self.volume * np.sin(2 * np.pi * self.f0 * t))
                )

        # concatenate pre-signal, data signal, and post-signal
        if self.pre_delay > 0 or self.post_delay > 0:
            fsk_signal = np.concatenate((pre_signal, fsk_signal, post_signal))

        # Normalize the signal for 100% volume
        if self.volume == 1.0:
            fsk_signal /= np.max(np.abs(fsk_signal))

        sci: AudioCodeInfo = AudioCodeInfo()
        sci.codec = AudioCodec.FSK
        sci.f1 = self.f1
        sci.f0 = self.f0
        sci.sample_rate = self.sample_rate
        sci.bit_duration = self.bit_duration
        sci.bit_count = c
        sci.volume = self.volume
        sci.duration = c * self.bit_duration + self.pre_delay + self.post_delay
        sci.pre_delay = self.pre_delay
        sci.post_delay = self.post_delay

        logger.debug(f"audio raw bits: count={c}, {sb}")
        logger.debug(f"audio duration: {sci.duration:.6f} seconds")
        return (fsk_signal, sci)

    def generate_nfe(self, data) -> (np.array, AudioCodeInfo):
        # Create NFE signal
        nfe_signal = np.array([])
        pre_signal = np.array([])
        post_signal = np.array([])

        # generate pre-signal if any
        if self.pre_delay > 0:
            pre_signal = self.generate_sin(self.pre_f, self.pre_delay)

        # generate post-signal if any
        if self.post_delay > 0:
            post_signal = self.generate_sin(self.post_f, self.post_delay)

        n = data.get_uint()
        c = int((self.f1 - self.f0) / self.nfe_df) + 1
        freq = self.f0 + (n % c) * self.nfe_df
        logger.debug(f" n={n}, c={c}, freq={freq}")
        nfe_signal = self.generate_sin(freq, self.nfe_duration)

        # concatenate pre-signal, data signal, and post-signal
        if self.pre_delay > 0 or self.post_delay > 0:
            nfe_signal = np.concatenate((pre_signal, nfe_signal, post_signal))

        # Normalize the signal for 100% volume
        if self.volume == 1.0:
            nfe_signal /= np.max(np.abs(nfe_signal))

        sci: AudioCodeInfo = AudioCodeInfo()
        sci.codec = AudioCodec.NFE
        sci.f1 = self.f1
        sci.f0 = self.f0
        sci.nfe_df = self.nfe_df
        sci.sample_rate = self.sample_rate
        sci.nfe_duration = self.nfe_duration
        sci.nfe_freq = freq
        sci.volume = self.volume
        sci.duration = self.nfe_duration + self.pre_delay + self.post_delay
        sci.pre_delay = self.pre_delay
        sci.post_delay = self.post_delay

        logger.debug(f"audio duration: {sci.duration:.6f} seconds")
        return nfe_signal, sci

    def generate(self, data) -> (np.array, AudioCodeInfo):
        if self.codec == AudioCodec.FSK:
            return self.generate_fsk(data)
        elif self.codec == AudioCodec.NFE:
            return self.generate_nfe(data)
        else:
            raise ValueError(f"Unsupported codec: {self.codec}")

    # play audio data with sounddevice
    def play_data_sd(self, data):
        ts = time.perf_counter()
        fsk_signal, sci = self.generate(data)
        ts = time.perf_counter() - ts
        logger.debug(f"generate time : {ts:.6f} seconds")

        ts = time.perf_counter()
        # Play the FSK signal
        sd.play(fsk_signal, samplerate=self.sample_rate)

        # Wait until audio is finished playing
        sd.wait()
        ts = time.perf_counter() - ts
        logger.debug(f"play time     : {ts:.6f} seconds")

    def save(self, data, filename):
        fsk_signal, sci = self.generate(data)

        # Save the signal to a WAV file
        write(filename, self.sample_rate, (fsk_signal * 32767).astype(np.int16))
        return sci

    def parse(self, filename):
        # Read the WAV file
        rate, data = wavfile.read(filename)

        # Check if audio is stereo and convert to mono if necessary
        if len(data.shape) > 1:
            data = data.mean(axis=1)

        # Calculate the number of samples for each bit duration
        samples_per_bit = int(self.sample_rate * self.bit_duration)

        # Prepare a list to hold the detected bits
        detected_bits = []

        # Analyze the audio in chunks
        for i in range(0, len(data), samples_per_bit):
            if i + samples_per_bit > len(data):
                break  # Avoid out of bounds

            # Extract the current chunk of audio
            chunk = data[i : i + samples_per_bit]  # noqa: E203

            # Perform FFT on the chunk
            fourier = np.fft.fft(chunk)
            frequencies = np.fft.fftfreq(len(fourier), 1 / rate)

            # Get the magnitudes and filter out positive frequencies
            magnitudes = np.abs(fourier)
            positive_frequencies = frequencies[: len(frequencies) // 2]
            positive_magnitudes = magnitudes[: len(magnitudes) // 2]

            # Find the peak frequency
            peak_freq = positive_frequencies[np.argmax(positive_magnitudes)]

            # Determine if the peak frequency corresponds to a '1' or '0'
            if abs(peak_freq - self.f1) < 50:  # Frequency for '1'
                detected_bits.append(1)
            elif abs(peak_freq - self.f0) < 50:  # Frequency for '0'
                detected_bits.append(0)

        dbg_bits: str = "".join([str(bit) for bit in detected_bits])
        logger.debug(f"detected bits : count={len(dbg_bits)}, {dbg_bits}")
        return bits_to_bytes(detected_bits)


######################################
# Public functions


def beep(duration: float = 2.0, async_: bool = False):
    logger.debug(f"beep(duration={duration})")
    play_audio("A", duration, async_)


def list_audio_devices():
    logger.debug("list_audio_devices()")

    logger.debug("[psychopy]")
    logger.debug(f"audioLib     : {prefs.hardware['audioLib']}")
    logger.debug(f"audioDevice  : {prefs.hardware['audioDevice']}")

    logger.debug("[sounddevice]")
    devices = sd.query_devices()  # Query all devices
    for i, device in enumerate(devices):
        logger.debug(f"device [{i}]  : {device['name']}")
    default_device = sd.default.device  # Get the current default input/output devices
    logger.debug(f"default in  : {default_device[0]}")
    logger.debug(f"default out : {default_device[1]}")

    logger.debug("[psytoolbox]")
    for i, device in enumerate(audio.get_devices()):
        logger.debug(f"device [{i}]  : {device}")

    logger.debug("[psychopy.backend_ptb]")
    # TODO: investigate why only single out device listed from
    # USB capture but default one is not shown
    # logger.debug(sound.backend_ptb.getDevices())


def _play_audio_psychopy(
    name: str,
    duration: float = None,
    volume: float = 0.8,
    sample_rate: int = 44100,
    async_: bool = False,
):
    logger.debug(
        f"_play_audio_psychopy(name={name}, duration={duration}, async_={async_})"
    )
    snd = None
    if duration:
        snd = sound.Sound(
            name, secs=duration, sampleRate=sample_rate, stereo=True, volume=volume
        )
    else:
        snd = sound.Sound(name, stereo=True, sampleRate=sample_rate, volume=volume)
    logger.debug(f"Play audio '{snd.sound}' with psychopy {prefs.hardware['audioLib']}")
    snd.play()
    logger.debug(
        f" sampleRate={snd.sampleRate}, duration={snd.getDuration()}, "
        f"volume={snd.getVolume()}"
    )
    if not async_:
        logger.debug("Waiting for audio to finish playing...")
        core.wait(snd.getDuration())
        logger.debug(f"Audio '{snd.sound}' has finished playing.")


def _play_audio_sd(
    name: str,
    duration: float = None,
    volume: float = 0.8,
    sample_rate: int = 44100,
    async_: bool = False,
):
    logger.debug(f"_play_audio_sd(name={name}, duration={duration}, async_={async_})")
    data = name

    if os.path.exists(name):
        rate, signal = read(name)
        logger.debug(f"Read audio file: {name}, rate={rate}")
        # Convert from int16 to float32
        # signal = signal.astype(np.float32) / 32767
        data = signal

    sd.play(data, samplerate=sample_rate)


def play_audio(
    name: str,
    duration: float = None,
    volume: float = 0.8,
    sample_rate: int = 44100,
    async_: bool = False,
):
    logger.debug(f"play_audio(name={name}, duration={duration}, async_={async_})")
    if (
        _audio_lib == AudioLib.PSYCHOPY_SOUNDDEVICE
        or _audio_lib == AudioLib.PSYCHOPY_PTB  # noqa: W503
    ):
        _play_audio_psychopy(name, duration, volume, sample_rate, async_)
    elif _audio_lib == AudioLib.SOUNDDEVICE:
        _play_audio_sd(name, duration, volume, sample_rate, async_)
    else:
        raise ValueError(f"Unsupported audio library: {_audio_lib}")


def save_audiocode(
    fname: str = None,
    code_uint16: int = None,
    code_uint32: int = None,
    code_uint64: int = None,
    code_str: str = None,
    code_bytes: bytes = None,
    codec: AudioCodec = AudioCodec.FSK,
    engine=None,
) -> (str, AudioCodeInfo):
    logger.debug(f"save_audiocode(fname={fname}...)")
    if not fname:
        fname = tempfile.mktemp(
            prefix=f"audiocode_{datetime.now().strftime('%Y%m%d_%H%M%S%f')}_",
            suffix=".wav",
        )

    data = DataMessage()
    if code_uint16 is not None:
        data.set_uint16(code_uint16)
    elif code_uint32 is not None:
        data.set_uint32(code_uint32)
    elif code_uint64 is not None:
        data.set_uint64(code_uint64)
    elif code_str:
        data.set_str(code_str)
    elif code_bytes:
        data.set_bytes(code_bytes)
    else:
        raise ValueError("No code data provided.")

    if not engine:
        engine = AudioCodeEngine(codec=codec)
    sci: AudioCodeInfo = engine.save(data, fname)
    logger.debug(f" -> {fname}")
    return (fname, sci)
